; File:		irTxSlowBiphaseATmega88A\main.asm
; Device:	ATmega88A
; Created:	21.10.2023
; Version:	2023-10-22
; Author:	Johannes Fechner

;.include <m88Adef.inc> ; implicitly included by Microchip Studio
.include "macros.asm"

; == General definitions ==
; Set the CPU clock frequency in Hz:
.equ F_CPU = 7_372_800

; == IR definitions ==
; Set some commands:
.equ IR_CMD_TEST0 = $D2
.equ IR_CMD_TEST1 = $79
; SlowBiphase modulation definitions:
.equ IR_WIDTH = 10 ; pulse or pause width in milliseconds

; == Hardware definitions ==
; Set switch connection:
.equ SWITCH_PIN = PINC
.equ SWITCH_PORT = PORTC
.equ SWITCH = 0
; Set OC0A = IR output (high active -> LED on at high state):
.equ OC0A_DDR = DDRD
.equ OC0A_PORT = PORTD
.equ OC0A = 6

; == Register definitions ==
; === Auxiliary registers ===
.def zero = r2
.def temp0 = r16
.def temp1 = r17
.def temp2 = r18

; === IR TX registers ===
.def ir_intrCntr = r19
.def ir_data = r20
.def ir_mask = r21

.org 0
	rjmp	reset
.org OC1Aaddr					; Timer1 Compare Match A interrupt
	rjmp	ir_intr
.org INT_VECTORS_SIZE

reset:
; Initialize stack pointer:
	ldi		temp0, high(RAMEND)
	xout	SPH, temp0
	ldi		temp0, low(RAMEND)
	xout	SPL, temp0
; Initialize general registers:
	clr		zero
; Set input pullups:
	xin		temp0, SWITCH_PORT
	ori		temp0, 1<<SWITCH
	xout	SWITCH_PORT, temp0
; Set output:
	sbi		OC0A_DDR, OC0A
; Initialize Timer0: 36 kHz carrier, output to OC0A:
.equ OCR0A_VAL = round(F_CPU/(2*36_000))-1
	ldi		temp0, OCR0A_VAL
	xout	OCR0A, temp0
; Initialize Timer1: 10 ms "tick", prescaling 1/1024:
.equ OCR1A_VAL = round(1.0*F_CPU*IR_WIDTH*50/(1024*1000))-1	; == 71 ; #### TEST: factor 50 in numerator ####
	ldi		temp0, high(OCR1A_VAL)
	xout	OCR1AH, temp0
	ldi		temp0, low(OCR1A_VAL)
	xout	OCR1AL, temp0
	ldi		temp0, 1<<WGM12							; CTC mode
	xout	TCCR1B, temp0
	ldi		temp0, 1<<OCIE1A						; enable Compare Match A interrupt
	xout	TIMSK1, temp0
; Enable interrupts:
	sei
; Main loop:
loop:
	xin		temp0, SWITCH_PIN
	sbrc	temp0, SWITCH
	rjmp	loop
	ldi		ir_data, IR_CMD_TEST0
	rcall	ir_transmit
	rcall	delay1s					; rudimentary switch debouncing
									; (for testing only)
	rjmp	loop

; == IR transmitter interrupt routine ==
ir_intr:
; Save temp0 and SREG:
	push	temp0
	xin		temp0, SREG
	push	temp0
; Test whether end of transmission is reached:
	cpi		ir_intrCntr, 17
	brsh	ir_intr_stop
; Test whether ir_intrCntr is even or odd:
	sbrc	ir_intrCntr, 0
	rjmp	ir_intr_odd
; Even, just toggle the carrier:
	xin		temp0, TCCR0B
	sbrc	temp0, CS00
	rjmp	ir_intr_carrierOff
	rjmp	ir_intr_carrierOn
ir_intr_stop:
; After last bit. Switch off carrier and stop interrupt timer:
	xin		temp0, TCCR1B
	andi	temp0, ~((1<<CS12) | (1<<CS10))		; stop timer
	xout	TCCR1B, temp0
	xout	TCNT1H, zero
	xout	TCNT1L, zero
	rjmp	ir_intr_carrierOff
ir_intr_odd:
; Odd ir_intrCntr, load the corresponding bit from ir_data:
	mov		temp0, ir_data
	and		temp0, ir_mask
	lsl		ir_mask					; prepare mask for next bit
	tst		temp0
	breq	ir_intr_data0
; Bit == 1, clear carrier if not already switched off:
	xin		temp0, TCCR0B
	sbrc	temp0, CS00
	rjmp	ir_intr_carrierOff
	rjmp	ir_intr_inc
ir_intr_data0:
; Bit == 0, set carrier if not already switched on:
	xin		temp0, TCCR0B
	sbrs	temp0, CS00
	rjmp	ir_intr_carrierOn
	rjmp	ir_intr_inc
ir_intr_carrierOff:
	xout	TCCR0B, zero						; stop Timer0
	ldi		temp0, 1<<WGM01						; disable I/O pin override
	xout	TCCR0A, temp0
	xout	TCNT0, zero
	cbi		OC0A_PORT, OC0A						; switch off IR LED
	rjmp	ir_intr_inc
ir_intr_carrierOn:
	sbi		OC0A_PORT, OC0A						; switch on IR LED
	ldi		temp0, (1<<COM0A0) | (1<<WGM01)		; toggle OC0A (I/O override), CTC mode
	xout	TCCR0A, temp0
	ldi		temp0, 1<<CS00						; start timer, no prescaling
	xout	TCCR0B, temp0
ir_intr_inc:
	inc		ir_intrCntr
ir_intr_end:
; Restore SREG and temp0:
	pop		temp0
	xout	SREG, temp0
	pop		temp0
	reti

; == Routine delay1s ==
; (for testing only)
delay1s:
	push	temp0
	push	temp1
	push	temp2
	ldi		temp0, F_CPU*10/607/1_000
delay1s_loop0:
	ldi		temp1, 201
delay1s_loop1:
	ldi		temp2, 100
delay1s_loop2:
	dec		temp2
	brne	delay1s_loop2
	dec		temp1
	brne	delay1s_loop1
	dec		temp0
	brne	delay1s_loop0
	pop		temp2
	pop		temp1
	pop		temp0
	ret

; == Routine ir_transmit ==
ir_transmit:
	push	temp0
	clr		ir_intrCntr
	ldi		ir_mask, 1
; Wait for previous transmission to have completed:
ir_transmit_wait:
	xin		temp0, TCCR1B
	sbrc	temp0, CS10
	rjmp	ir_transmit_wait
; Start Timer1:
	xin		temp0, TCCR1B
	ori		temp0, (1<<CS12) | (1<<CS10)			; prescaling 1/1024
	xout	TCCR1B, temp0
	pop		temp0
	ret
